#include "Arduino.h"
#include "BNC_CodingDrone_def.h"
#include "BNC_CodingDrone_structs.h"
#include "BNC_CodingDrone.h"
#include "BNC_CodingDrone_Sensors.h"

#define UPDATE_INTERVAL 20000    // 40hz update rate (20hz LPF on acc)
#define BARO_TAB_SIZE   21
#define ACC_Z_DEADBAND (ACC_1G>>5) // was 40 instead of 32 now

 /**************************************************************************************/
  /********                              Gyro filters                ********************/
  /**************************************************************************************/

/*********************    Lowpass filter for some gyros    ****************************/
  /* ITG3200 & ITG3205 Low pass filter setting. In case you cannot eliminate all vibrations to the Gyro, you can try
	 to decrease the LPF frequency, only one step per try. As soon as twitching gone, stick with that setting.
	 It will not help on feedback wobbles, so change only when copter is randomly twiching and all dampening and
	 balancing options ran out. Uncomment only one option!
	 IMPORTANT! Change low pass filter setting changes PID behaviour, so retune your PID's after changing LPF.
	 available for ITG3050, ITG3200, MPU3050, MPU6050*/
  //#define GYRO_LPF_256HZ     // This is the default setting, no need to uncomment, just for reference
  #define GYRO_LPF_188HZ
  //#define GYRO_LPF_98HZ
  //#define GYRO_LPF_42HZ
  //#define GYRO_LPF_20HZ
  //#define GYRO_LPF_10HZ
  //#define GYRO_LPF_5HZ       // Use this only in extreme cases, rather change motors and/or props -- setting not available on ITG3200

/******                Gyro smoothing    **********************************/
  /* GYRO_SMOOTHING. In case you cannot reduce vibrations _and_ _after_ you have tried the low pass filter options, you
	 may try this gyro smoothing via averaging. Not suitable for multicopters!
	 Good results for helicopter, airplanes and flying wings (foamies) with lots of vibrations.*/
  //#define GYRO_SMOOTHING {20, 20, 3}    // (*) separate averaging ranges for roll, pitch, yaw

/************************    Moving Average Gyros    **********************************/
  //#define MMGYRO 10                      // (*) Active Moving Average Function for Gyros
  //#define MMGYROVECTORLENGTH 15          // Length of Moving Average Vector (maximum value for tunable MMGYRO
  /* Moving Average ServoGimbal Signal Output */
  //#define MMSERVOGIMBAL                  // Active Output Moving Average Function for Servos Gimbal
  //#define MMSERVOGIMBALVECTORLENGHT 32   // Lenght of Moving Average Vector



#define applyDeadband(value, deadband)  \
  if(abs(value) < deadband) {           \
    value = 0;                          \
  } else if(value > 0){                 \
    value -= deadband;                  \
  } else if(value < 0){                 \
    value += deadband;                  \
  }

// ***********************************
// Air Pressure Sensor Data Variables
// ***********************************
int32_t baroPressure;
int16_t baroTemperature;
int32_t baroPressureSum;

// **************
// 자이로 가속도 센서 변수 및 객체
// **************

int16_t gyroZero[3] = {0,0,0};
float gyroZero_Yaw = 0.0f;
uint16_t calibratingA = 0;  // the calibration is done in the main loop. Calibrating decreases at each cycle down to 0, then we enter in a normal mode.
uint16_t calibratingB = 0;  // baro calibration = get new ground pressure value
uint16_t calibratingG = 0;


imu_t imu;
alt_t alt;
att_t att;

static void Device_Mag_getADC();
static void init_AirPressure();
static void Mag_init();
static void ACC_init();

int16_t  i2c_errors_count = 0;
int16_t imu_yaw_offset = 0;

// ************************************************************************************************************
// board orientation and setup
// ************************************************************************************************************

//ITG3200 / ITG3205 / ITG3050 / MPU6050 / MPU3050 Gyro LPF setting
#if defined(GYRO_LPF_256HZ) || defined(GYRO_LPF_188HZ) || defined(GYRO_LPF_98HZ) || defined(GYRO_LPF_42HZ) || defined(GYRO_LPF_20HZ) || defined(GYRO_LPF_10HZ) || defined(GYRO_LPF_5HZ)
  #if defined(GYRO_LPF_256HZ)
    #define GYRO_DLPF_CFG   0
  #endif
  #if defined(GYRO_LPF_188HZ)
    #define GYRO_DLPF_CFG   1
  #endif
  #if defined(GYRO_LPF_98HZ)
    #define GYRO_DLPF_CFG   2
  #endif
  #if defined(GYRO_LPF_42HZ)
    #define GYRO_DLPF_CFG   3
  #endif
  #if defined(GYRO_LPF_20HZ)
    #define GYRO_DLPF_CFG   4
  #endif
  #if defined(GYRO_LPF_10HZ)
    #define GYRO_DLPF_CFG   5
  #endif
  #if defined(GYRO_LPF_5HZ)
    #define GYRO_DLPF_CFG   6
  #endif
#else
    #define GYRO_DLPF_CFG   0 //Default settings LPF 256Hz/8000Hz sample
#endif

static uint8_t rawADC[6];
#if defined(WMP)
static uint32_t neutralizeTime = 0;
#endif

// ************************************************************************************************************
// I2C general functions
// ************************************************************************************************************

void i2c_init(void) {
  #if defined(INTERNAL_I2C_PULLUPS)
    I2C_PULLUPS_ENABLE
  #else
    I2C_PULLUPS_DISABLE
  #endif
  TWSR = 0;                                    // no prescaler => prescaler = 1
  TWBR = ((F_CPU / 400000) - 16) / 2;          // set the I2C clock rate to 400kHz
  TWCR = 1<<TWEN;                              // enable twi module, no interrupt
  i2c_errors_count = 0;
}

void __attribute__ ((noinline)) waitTransmissionI2C(uint8_t twcr) {
  TWCR = twcr;
  uint8_t count = 255;
  while (!(TWCR & (1<<TWINT))) {
    count--;
    if (count==0) {              //we are in a blocking state => we don't insist
      TWCR = 0;                  //and we force a reset on TWINT register
      #if defined(WMP)
      neutralizeTime = micros(); //we take a timestamp here to neutralize the value during a short delay
      #endif
      i2c_errors_count++;
      break;
    }
  }
}

void i2c_rep_start(uint8_t address) {
  waitTransmissionI2C((1<<TWINT) | (1<<TWSTA) | (1<<TWEN)); // send REPEAT START condition and wait until transmission completed
  TWDR = address;                                           // send device address
  waitTransmissionI2C((1<<TWINT) | (1<<TWEN));              // wail until transmission completed
}

void i2c_stop(void) {
  TWCR = (1 << TWINT) | (1 << TWEN) | (1 << TWSTO);
  //  while(TWCR & (1<<TWSTO));                // <- can produce a blocking state with some WMP clones
}

void i2c_write(uint8_t data ) {
  TWDR = data;                                 // send data to the previously addressed device
  waitTransmissionI2C((1<<TWINT) | (1<<TWEN));
}

uint8_t i2c_readAck() {
  waitTransmissionI2C((1<<TWINT) | (1<<TWEN) | (1<<TWEA));
  return TWDR;
}

uint8_t i2c_readNak() {
  waitTransmissionI2C((1<<TWINT) | (1<<TWEN));
  uint8_t r = TWDR;
  i2c_stop();
  return r;
}

void i2c_read_reg_to_buf(uint8_t add, uint8_t reg, uint8_t *buf, uint8_t size) {
  i2c_rep_start(add<<1); // I2C write direction
  i2c_write(reg);        // register selection
  i2c_rep_start((add<<1) | 1);  // I2C read direction
  uint8_t *b = buf;
  while (--size) *b++ = i2c_readAck(); // acknowledge all but the final byte
  *b = i2c_readNak();
}

void i2c_getSixRawADC(uint8_t add, uint8_t reg) {
  i2c_read_reg_to_buf(add, reg, rawADC, 6);
}

void i2c_writeReg(uint8_t add, uint8_t reg, uint8_t val) {
  i2c_rep_start(add<<1); // I2C write direction
  i2c_write(reg);        // register selection
  i2c_write(val);        // value to write in register
  i2c_stop();
}

uint8_t i2c_readReg(uint8_t add, uint8_t reg) {
  uint8_t val;
  i2c_read_reg_to_buf(add, reg, &val, 1);
  return val;
}

// ****************
// GYRO common part
// ****************
void GYRO_Common() {
  static int16_t previousGyroADC[3] = {0,0,0};
  static int32_t g[3];
  uint8_t axis, tilt=0;

  #if defined MMGYRO       
    // Moving Average Gyros by Magnetron1
    //---------------------------------------------------
    static int16_t mediaMobileGyroADC[3][MMGYROVECTORLENGTH];
    static int32_t mediaMobileGyroADCSum[3];
    static uint8_t mediaMobileGyroIDX;
    //---------------------------------------------------
  #endif

  if (calibratingG>0) {
    for (axis = 0; axis < 3; axis++) {
      if (calibratingG == 128) { // Reset g[axis] at start of calibration
        g[axis]=0;
      }
      g[axis] += imu.gyroADC[axis]; // Sum up 512 readings
      if (calibratingG == 1) {
        //SET_ALARM_BUZZER(ALRM_FAC_CONFIRM, ALRM_LVL_CONFIRM_ELSE);
		gyroZero[axis]=g[axis]/128;
		//if(axis == 2) gyroZero_Yaw = (float)g[axis]/128;
		att.angle_Yaw = 0;
      }
    }
      calibratingG--;
  }

  #ifdef MMGYRO       
  mediaMobileGyroIDX = ++mediaMobileGyroIDX % conf.mmgyro;
  for (axis = 0; axis < 3; axis++) {
    imu.gyroADC[axis]  -= gyroZero[axis];
    mediaMobileGyroADCSum[axis] -= mediaMobileGyroADC[axis][mediaMobileGyroIDX];
    //anti gyro glitch, limit the variation between two consecutive readings
    mediaMobileGyroADC[axis][mediaMobileGyroIDX] = constrain(imu.gyroADC[axis],previousGyroADC[axis]-800,previousGyroADC[axis]+800);
    mediaMobileGyroADCSum[axis] += mediaMobileGyroADC[axis][mediaMobileGyroIDX];
    imu.gyroADC[axis] = mediaMobileGyroADCSum[axis] / conf.mmgyro;
    previousGyroADC[axis] = imu.gyroADC[axis];
  }
  #else
  for (axis = 0; axis < 3; axis++) {
    imu.gyroADC[axis]  -= gyroZero[axis];
    //anti gyro glitch, limit the variation between two consecutive readings
    imu.gyroADC[axis] = constrain(imu.gyroADC[axis],previousGyroADC[axis]-800,previousGyroADC[axis]+800);
    previousGyroADC[axis] = imu.gyroADC[axis];
  }
  #endif    

}

// ****************
// ACC common part
// ****************
void ACC_Common() {
  static int32_t a[3];
  if (calibratingA>0) {
    calibratingA--;
    for (uint8_t axis = 0; axis < 3; axis++) {
      if (calibratingA == 127) a[axis]=0;   // Reset a[axis] at start of calibration
      a[axis] +=imu.accADC[axis];           // Sum up 512 readings
      global_conf.accZero[axis] = a[axis]>>7; // Calculate average, only the last itteration where (calibratingA == 0) is relevant
    }
    if (calibratingA == 0) {
      global_conf.accZero[YAW] -= ACC_1G;   // shift Z down by ACC_1G and store values in EEPROM at end of calibration
      //conf.angleTrim[ROLL]   = ROLL_TRIM;
      //conf.angleTrim[PITCH]  = PITCH_TRIM;
      //writeGlobalSet(1); // write accZero in EEPROM
    }
  }
  #if defined(INFLIGHT_ACC_CALIBRATION)
      static int32_t b[3];
      static int16_t accZero_saved[3]  = {0,0,0};
      static int16_t  angleTrim_saved[2] = {0, 0};
      //Saving old zeropoints before measurement
      if (InflightcalibratingA==50) {
         accZero_saved[ROLL]  = global_conf.accZero[ROLL] ;
         accZero_saved[PITCH] = global_conf.accZero[PITCH];
         accZero_saved[YAW]   = global_conf.accZero[YAW] ;
         angleTrim_saved[ROLL]  = conf.angleTrim[ROLL] ;
         angleTrim_saved[PITCH] = conf.angleTrim[PITCH] ;
      }
      if (InflightcalibratingA>0) {
        for (uint8_t axis = 0; axis < 3; axis++) {
          // Reset a[axis] at start of calibration
          if (InflightcalibratingA == 50) b[axis]=0;
          // Sum up 50 readings
          b[axis] +=imu.accADC[axis];
          // Clear global variables for next reading
          imu.accADC[axis]=0;
          global_conf.accZero[axis]=0;
        }
        //all values are measured
        if (InflightcalibratingA == 1) {
          AccInflightCalibrationActive = 0;
          AccInflightCalibrationMeasurementDone = 1;
          //SET_ALARM_BUZZER(ALRM_FAC_CONFIRM, ALRM_LVL_CONFIRM_1);     //buzzer for indicatiing the end of calibration
          // recover saved values to maintain current flight behavior until new values are transferred
          global_conf.accZero[ROLL]  = accZero_saved[ROLL] ;
          global_conf.accZero[PITCH] = accZero_saved[PITCH];
          global_conf.accZero[YAW]   = accZero_saved[YAW] ;
          conf.angleTrim[ROLL]  = angleTrim_saved[ROLL] ;
          conf.angleTrim[PITCH] = angleTrim_saved[PITCH] ;
        }
        InflightcalibratingA--;
      }
      // Calculate average, shift Z down by ACC_1G and store values in EEPROM at end of calibration
      if (AccInflightCalibrationSavetoEEProm == 1){  //the copter is landed, disarmed and the combo has been done again
        AccInflightCalibrationSavetoEEProm = 0;
        global_conf.accZero[ROLL]  = b[ROLL]/50;
        global_conf.accZero[PITCH] = b[PITCH]/50;
        global_conf.accZero[YAW]   = b[YAW]/50-ACC_1G;
        conf.angleTrim[ROLL]   = ROLL_TRIM;
        conf.angleTrim[PITCH]  = PITCH_TRIM;
        writeGlobalSet(1); // write accZero in EEPROM
      }
  #endif
  imu.accADC[ROLL]  -=  global_conf.accZero[ROLL];
  imu.accADC[PITCH] -=  global_conf.accZero[PITCH];
  imu.accADC[YAW]   -=  global_conf.accZero[YAW];

  #if defined(SENSORS_TILT_45DEG_LEFT)
    int16_t temp = ((imu.accADC[PITCH] - imu.accADC[ROLL] )*7)/10;
    imu.accADC[ROLL] = ((imu.accADC[ROLL] + imu.accADC[PITCH])*7)/10;
    imu.accADC[PITCH] = temp;
  #endif
  #if defined(SENSORS_TILT_45DEG_RIGHT)
    int16_t temp = ((imu.accADC[PITCH] + imu.accADC[ROLL] )*7)/10;
    imu.accADC[ROLL] = ((imu.accADC[ROLL] - imu.accADC[PITCH])*7)/10;
    imu.accADC[PITCH] = temp;
  #endif
}

// ************************************************************************************************************
// BARO section
// ************************************************************************************************************

static void Baro_Common() {
  static int32_t baroHistTab[BARO_TAB_SIZE];
  static uint8_t baroHistIdx;

  uint8_t indexplus1 = (baroHistIdx + 1);
  if (indexplus1 == BARO_TAB_SIZE) indexplus1 = 0;
  baroHistTab[baroHistIdx] = baroPressure;
  baroPressureSum += baroHistTab[baroHistIdx];
  baroPressureSum -= baroHistTab[indexplus1];
  baroHistIdx = indexplus1;
}

// ************************************************************************************************************
// I2C Barometer BNC_PRESSURE_SENSOR
// ************************************************************************************************************
// I2C adress: 0x76 (7bit)
// principle:
//  1) read the calibration register (only once at the initialization)
//  2) read uncompensated temperature (not mandatory at every cycle)
//  3) read uncompensated pressure
//  4) raw temp + raw pressure => calculation of the adjusted pressure
//  the following code uses the maximum precision setting
// ************************************************************************************************************

#define PS_ADDRESS 0x76
#define PS_CHIPID  0x58

//???量速率(sample/sec),Background 模式使用
#define  PM_RATE_1          (0<<4)      //1 measurements pr. sec.
#define  PM_RATE_2          (1<<4)      //2 measurements pr. sec.
#define  PM_RATE_4          (2<<4)      //4 measurements pr. sec.           
#define  PM_RATE_8          (3<<4)      //8 measurements pr. sec.
#define  PM_RATE_16         (4<<4)      //16 measurements pr. sec.
#define  PM_RATE_32         (5<<4)      //32 measurements pr. sec.
#define  PM_RATE_64         (6<<4)      //64 measurements pr. sec.
#define  PM_RATE_128        (7<<4)      //128 measurements pr. sec.

//??重采?速率(times),Background 模式使用
#define PM_PRC_1            0       //Sigle         kP=524288   ,3.6ms
#define PM_PRC_2            1       //2 times       kP=1572864  ,5.2ms
#define PM_PRC_4            2       //4 times       kP=3670016  ,8.4ms
#define PM_PRC_8            3       //8 times       kP=7864320  ,14.8ms
#define PM_PRC_16           4       //16 times      kP=253952   ,27.6ms
#define PM_PRC_32           5       //32 times      kP=516096   ,53.2ms
#define PM_PRC_64           6       //64 times      kP=1040384  ,104.4ms
#define PM_PRC_128          7       //128 times     kP=2088960  ,206.8ms

//?度?量速率(sample/sec),Background 模式使用
#define  TMP_RATE_1         (0<<4)      //1 measurements pr. sec.
#define  TMP_RATE_2         (1<<4)      //2 measurements pr. sec.
#define  TMP_RATE_4         (2<<4)      //4 measurements pr. sec.           
#define  TMP_RATE_8         (3<<4)      //8 measurements pr. sec.
#define  TMP_RATE_16        (4<<4)      //16 measurements pr. sec.
#define  TMP_RATE_32        (5<<4)      //32 measurements pr. sec.
#define  TMP_RATE_64        (6<<4)      //64 measurements pr. sec.
#define  TMP_RATE_128       (7<<4)      //128 measurements pr. sec.

//?度重采?速率(times),Background 模式使用
#define TMP_PRC_1           0       //Sigle
#define TMP_PRC_2           1       //2 times
#define TMP_PRC_4           2       //4 times
#define TMP_PRC_8           3       //8 times
#define TMP_PRC_16          4       //16 times
#define TMP_PRC_32          5       //32 times
#define TMP_PRC_64          6       //64 times
#define TMP_PRC_128         7       //128 times

//SPL06_MEAS_CFG
#define MEAS_COEF_RDY       0x80
#define MEAS_SENSOR_RDY     0x40        //?感器初始化完成
#define MEAS_TMP_RDY        0x20        //有新的?度?据
#define MEAS_PRS_RDY        0x10        //有新的???据

#define MEAS_CTRL_Standby               0x00    //空?模式
#define MEAS_CTRL_PressMeasure          0x01    //?次???量
#define MEAS_CTRL_TempMeasure           0x02    //?次?度?量
#define MEAS_CTRL_ContinuousPress       0x05    //?????量
#define MEAS_CTRL_ContinuousTemp        0x06    //???度?量
#define MEAS_CTRL_ContinuousPressTemp   0x07    //?????度?量


//FIFO_STS
#define FIFO_FULL     0x02
#define FIFO_EMPTY    0x01

//INT_STS
#define INT_FIFO_FULL     0x04
#define INT_TMP           0x02
#define INT_PRS           0x01

//CFG_REG
#define INT_SDO_PIN_CFG    0x80        //???量配置
#define INT_FIFO_FULL_CFG    0x40        //?度?量配置
#define PSR_CFG    0x20        //???量配置
#define TMP_CFG    0x10        //?度?量配置
#define CFG_T_SHIFT   0x08    //oversampling times>8?必?使用
#define CFG_P_SHIFT   0x04
#define FIFO_EN   0x02
#define SPI_MODE   0x01

/*=========================================================================
    REGISTERS
    -----------------------------------------------------------------------*/
enum
{
  PS_REGISTER_PSR_B2              = 0x00,
  PS_REGISTER_PSR_B1              = 0x01,
  PS_REGISTER_PSR_B0              = 0x02,

  PS_REGISTER_TMP_B2              = 0x03,
  PS_REGISTER_TMP_B1              = 0x04,
  PS_REGISTER_TMP_B0              = 0x05,

  PS_REGISTER_PSR_CFG              = 0x06,
  PS_REGISTER_TMP_CFG              = 0x07,

  PS_REGISTER_MEAS_CFG             = 0x08,
  PS_REGISTER_CFG_REG              = 0x09,
  PS_REGISTER_INS_STS              = 0x0A,
  PS_REGISTER_FIFO_STS             = 0x0B,

  PS_REGISTER_SOFTRESET            = 0x0C,
  PS_REGISTER_CHIPID               = 0x0D,

  PS_REGISTER_DIG_C0              = 0x10,
  PS_REGISTER_DIG_C0_C1           = 0x11,
  PS_REGISTER_DIG_C1              = 0x12,
  PS_REGISTER_DIG_C00_1           = 0x13,
  PS_REGISTER_DIG_C00_2           = 0x14,
  PS_REGISTER_DIG_C00_3_C10_1     = 0x15,
  PS_REGISTER_DIG_C10_2           = 0x16,
  PS_REGISTER_DIG_C10_3           = 0x17,
  PS_REGISTER_DIG_C01_1           = 0x18,
  PS_REGISTER_DIG_C01_2           = 0x19,
  PS_REGISTER_DIG_C11_1           = 0x1A,
  PS_REGISTER_DIG_C11_2           = 0x1B,
  PS_REGISTER_DIG_C20_1           = 0x1C,
  PS_REGISTER_DIG_C20_2           = 0x1D,
  PS_REGISTER_DIG_C21_1           = 0x1E,
  PS_REGISTER_DIG_C21_2           = 0x1F,
  PS_REGISTER_DIG_C30_1           = 0x20,
  PS_REGISTER_DIG_C30_2           = 0x21,

  PS_STATUS_COEF_RDY              = 0b10000000,
  PS_STATUS_SENSOR_RDY            = 0b01000000,
  PS_STATUS_TMP_RDY               = 0b00100000,
  PS_STATUS_PRS_RDY               = 0b00010000,
  PS_STATUS_TMP_PRS_RDY           = 0b00110000,
  PS_TMP_MEAS_CTRL                = 0b11000010,
  PS_PRS_MEAS_CTRL                = 0b11000001,

};

/*=========================================================================*/

static struct {
  // sensor registers from the PS datasheet
  int16_t dig_C0;
  int16_t dig_C1;
  int32_t dig_C00;
  int32_t dig_C10;
  int16_t dig_C01;
  int16_t dig_C11;
  int16_t dig_C20;
  int16_t dig_C21;
  int16_t dig_C30;
  uint8_t buffer[18];

  uint8_t t_OSR;
  uint8_t p_OSR;

  float Kt;
  float Kp;

  uint8_t T_buffer[3];
  uint8_t P_buffer[3];

  int32_t T_raw;
  int32_t P_raw;

  float T_raw_sc;
  float P_raw_sc;
  float T_comp;
  float T_comp_LPF;
  float P_comp;
  float P_comp_LPF;


  uint8_t Buff_status;
  uint8_t INT_status;
  uint8_t FIFO_status;

  uint8_t state;
  uint32_t deadline;

  uint8_t comm_Error_cout;
} PS_ctx;

/* transform a series of bytes from big endian to little
   endian and vice versa. */
   
void swap_endianness(void *buf, size_t size) {
  /* we swap in-place, so we only have to
  * place _one_ element on a temporary tray
  */
  uint8_t tray;
  uint8_t *from;
  uint8_t *to;
  /* keep swapping until the pointers have assed each other */
  for (from = (uint8_t*)buf, to = &from[size-1]; from < to; from++, to--) {
    tray = *from;
    *from = *to;
    *to = tray;
  }
}

void i2c_PS_readCalibration() {
  delay(10);
  //read calibration data in one go
  i2c_read_reg_to_buf(PS_ADDRESS, PS_REGISTER_DIG_C0, PS_ctx.buffer, 18);

  PS_ctx.dig_C0 = ((int16_t)(PS_ctx.buffer[1] & 0xF0) >> 4) + ((int16_t)PS_ctx.buffer[0] << 4);
  PS_ctx.dig_C0 = (PS_ctx.dig_C0 & 0x0800) ? (0xF000 | PS_ctx.dig_C0) : PS_ctx.dig_C0;

  PS_ctx.dig_C1 = ((int16_t)(PS_ctx.buffer[1] & 0x0F) << 8) + (int16_t)PS_ctx.buffer[2];
  PS_ctx.dig_C1 = (PS_ctx.dig_C1 & 0x0800) ? (0xF000 | PS_ctx.dig_C1) : PS_ctx.dig_C1;

  PS_ctx.dig_C00 = ((int32_t)PS_ctx.buffer[5] >> 4) + ((int32_t)PS_ctx.buffer[4] << 4) + ((int32_t)PS_ctx.buffer[3] << 12);
  PS_ctx.dig_C00 = (PS_ctx.dig_C00 & 0x080000) ? (0xFFF00000 | PS_ctx.dig_C00) : PS_ctx.dig_C00;

  PS_ctx.dig_C10 = ((int32_t)(PS_ctx.buffer[5] & 0x0F) << 16) + ((int32_t)PS_ctx.buffer[6] << 8) + (int32_t)PS_ctx.buffer[7];
  PS_ctx.dig_C10 = (PS_ctx.dig_C10 & 0x080000) ? (0xFFF00000 | PS_ctx.dig_C10) : PS_ctx.dig_C10;

  PS_ctx.dig_C01 = ((int16_t)PS_ctx.buffer[8] << 8) + (int16_t)PS_ctx.buffer[9];

  PS_ctx.dig_C11 = ((int16_t)PS_ctx.buffer[10] << 8) + (int16_t)PS_ctx.buffer[11];
  PS_ctx.dig_C11 = (PS_ctx.dig_C11 & 0x0800) ? (0xF000 | PS_ctx.dig_C11) : PS_ctx.dig_C11;

  PS_ctx.dig_C20 = ((int16_t)PS_ctx.buffer[12] << 8) + (int16_t)PS_ctx.buffer[13];
  PS_ctx.dig_C20 = (PS_ctx.dig_C20 & 0x0800) ? (0xF000 | PS_ctx.dig_C20) : PS_ctx.dig_C20;

  PS_ctx.dig_C21 = ((int16_t)PS_ctx.buffer[14] << 8) + (int16_t)PS_ctx.buffer[15];
  PS_ctx.dig_C21 = (PS_ctx.dig_C21 & 0x0800) ? (0xF000 | PS_ctx.dig_C21) : PS_ctx.dig_C21;

  PS_ctx.dig_C30 = ((int16_t)PS_ctx.buffer[16] << 8) + (int16_t)PS_ctx.buffer[17];
  PS_ctx.dig_C30 = (PS_ctx.dig_C30 & 0x0800) ? (0xF000 | PS_ctx.dig_C30) : PS_ctx.dig_C30;
}

void i2c_PS_SW_Reset(void) {
  i2c_writeReg(PS_ADDRESS,PS_REGISTER_SOFTRESET,0b10001001);
}

void i2c_PS_ReadStatus() {
  i2c_rep_start(PS_ADDRESS<<1); // I2C write direction
  i2c_write(PS_REGISTER_MEAS_CFG);        // register selection
  i2c_stop();
  i2c_rep_start((PS_ADDRESS<<1) | 1);//I2C read direction => 1
  PS_ctx.Buff_status = i2c_readNak();
}

void i2c_PS_MEAS_Start() {
  i2c_writeReg(PS_ADDRESS, PS_REGISTER_MEAS_CFG, MEAS_CTRL_ContinuousPressTemp);
  //i2c_rep_start((PS_ADDRESS<<1) | 1);//I2C read direction => 1
  //PS_ctx.Buff_status = i2c_readNak();
}

void i2c_PS_UP_Start() {
  i2c_writeReg(PS_ADDRESS,PS_REGISTER_MEAS_CFG,PS_PRS_MEAS_CTRL);
  //i2c_rep_start((PS_ADDRESS<<1) | 1);//I2C read direction => 1
  //PS_ctx.Buff_status = i2c_readNak();
}

void i2c_PS_UT_Start() {
  i2c_writeReg(PS_ADDRESS,PS_REGISTER_MEAS_CFG,PS_TMP_MEAS_CTRL);
  //i2c_rep_start((PS_ADDRESS<<1) | 1);//I2C read direction => 1
  //PS_ctx.Buff_status = i2c_readNak();
}

void i2c_PS_INT_Status() {
  i2c_rep_start(PS_ADDRESS<<1); // I2C write direction
  i2c_write(PS_REGISTER_INS_STS);        // register selection
  i2c_stop();
  i2c_rep_start((PS_ADDRESS<<1) | 1);//I2C read direction => 1
  PS_ctx.INT_status = i2c_readNak();
}

// read uncompensated pressure value: read result bytes
// the datasheet suggests a delay of 25.5 ms (oversampling settings 3) after the send command
void i2c_PS_UP_Read_Command() {
  i2c_rep_start(PS_ADDRESS<<1); // I2C write direction
  i2c_write(PS_REGISTER_PSR_B2);        // register selection
  i2c_stop();
}

void i2c_PS_FIFO_Status() {
  i2c_rep_start(PS_ADDRESS << 1); // I2C write direction
  i2c_write(PS_REGISTER_FIFO_STS);        // register selection
  i2c_stop();
  i2c_rep_start((PS_ADDRESS << 1) | 1); //I2C read direction => 1
  PS_ctx.FIFO_status = i2c_readNak();
}

void i2c_PS_UP_Read () {
  i2c_rep_start((PS_ADDRESS<<1) | 1);//I2C read direction => 1
  PS_ctx.P_buffer[0] = i2c_readAck();
  PS_ctx.P_buffer[1] = i2c_readAck();
  PS_ctx.P_buffer[2] = i2c_readNak();
}

// read uncompensated temperature value: read result bytes
// the datasheet suggests a delay of 4.5 ms after the send command
void i2c_PS_UT_Read_Command() {
  i2c_rep_start(PS_ADDRESS<<1); // I2C write direction
  i2c_write(PS_REGISTER_TMP_B2);        // register selection
  i2c_stop();
}

void i2c_PS_UT_Read() {
  i2c_rep_start((PS_ADDRESS<<1) | 1);//I2C read direction => 1
  PS_ctx.T_buffer[0] = i2c_readAck();
  PS_ctx.T_buffer[1] = i2c_readAck();
  PS_ctx.T_buffer[2] = i2c_readNak();
}

void i2c_PS_Temp_Calculate() {

  // Temperature calculations
  PS_ctx.T_raw = ((uint32_t)PS_ctx.T_buffer[0] << 16) + ((uint32_t)PS_ctx.T_buffer[1] << 8) + PS_ctx.T_buffer[2];
  PS_ctx.T_raw_sc = (float)PS_ctx.T_raw / PS_ctx.Kt ;
  //PS_ctx.T_comp = (PS_ctx.dig_C0 * 0.5f + PS_ctx.T_raw_sc * PS_ctx.dig_C1);
  
  //if(PS_ctx.T_comp_LPF == 0) PS_ctx.T_comp_LPF = PS_ctx.T_comp;
  //else PS_ctx.T_comp_LPF = (PS_ctx.T_comp_LPF*0.95 + PS_ctx.T_comp*0.05);
  //baroTemperature = PS_ctx.T_comp;

  //(int32_t)(PS_ctx.P_comp/10);
  //baroPressure = (int32_t)(PS_ctx.P_comp);
}
  
void i2c_PS_Pressure_Calculate() {

  // Temperature calculations
  float qua2, qua3;
  PS_ctx.P_raw = ((uint32_t)PS_ctx.P_buffer[0] << 16) + ((uint32_t)PS_ctx.P_buffer[1] << 8) + PS_ctx.P_buffer[2];
  PS_ctx.P_raw = (PS_ctx.P_raw & 0x800000) ? (0xFF000000 | PS_ctx.P_raw) : PS_ctx.P_raw;
  PS_ctx.P_raw_sc = (float)PS_ctx.P_raw / PS_ctx.Kp;
  // PS_ctx.P_buffer = ((int64_t)PS_ctx.dig_C00
  // + (float)PS_ctx.P_raw * ((int64_t)PS_ctx.dig_C10 + (float)PS_ctx.P_raw * ((int64_t)PS_ctx.dig_C20 / (float)PS_ctx.Kp
  // + (float)PS_ctx.P_raw * (int64_t)PS_ctx.dig_C30 / (float)PS_ctx.Kp)) / (float)PS_ctx.Kp
  // + (float)PS_ctx.T_raw * (int64_t)PS_ctx.dig_C01 / (float)PS_ctx.Kt
  // + PS_ctx.P_raw_sc * ((int64_t)PS_ctx.dig_C11
  // + (float)PS_ctx.P_raw * (int64_t)PS_ctx.dig_C21 / (float)PS_ctx.Kp) / (float)PS_ctx.Kp * (float)PS_ctx.T_raw / (float)PS_ctx.Kt);
  // PS_ctx.P_comp = (float)(PS_ctx.P_buffer);

  qua2 = PS_ctx.dig_C10  + PS_ctx.P_raw_sc * (PS_ctx.dig_C20  + PS_ctx.P_raw_sc * PS_ctx.dig_C30 );
  qua3 = PS_ctx.T_raw_sc * PS_ctx.P_raw_sc * (PS_ctx.dig_C11  + PS_ctx.P_raw_sc * PS_ctx.dig_C21 );
  PS_ctx.P_comp = PS_ctx.dig_C00  + PS_ctx.P_raw_sc * qua2 + PS_ctx.T_raw_sc * PS_ctx.dig_C01 + qua3;
  //if(PS_ctx.P_comp_LPF == 0) PS_ctx.P_comp_LPF = PS_ctx.P_comp;
  //else PS_ctx.P_comp_LPF = (PS_ctx.P_comp_LPF*0.985 + PS_ctx.P_comp*0.015);
  //(int32_t)(PS_ctx.P_comp/10);
  //baroPressure = (int32_t)(PS_ctx.P_comp);
}

void i2c_PS_Calculate() {
  
  // Temperature calculations
  float qua2, qua3;
  PS_ctx.T_raw = ((uint32_t)PS_ctx.T_buffer[0]<<16) + ((uint32_t)PS_ctx.T_buffer[1]<<8) + PS_ctx.T_buffer[2];
  PS_ctx.T_raw_sc = (float)PS_ctx.T_raw / PS_ctx.Kt ;
  PS_ctx.T_comp = (PS_ctx.dig_C0 * 0.5f + PS_ctx.T_raw_sc * PS_ctx.dig_C1);
  //baroTemperature = PS_ctx.T_comp;

  // Pressure calculations
  //PS_ctx.P_raw = (int32_t)((PS_ctx.P_buffer[0]<<16) + (PS_ctx.P_buffer[1]<<8) + PS_ctx.P_buffer[2]);
  PS_ctx.P_raw = ((uint32_t)PS_ctx.P_buffer[0]<<16) + ((uint32_t)PS_ctx.P_buffer[1]<<8) + PS_ctx.P_buffer[2];
  PS_ctx.P_raw = (PS_ctx.P_raw&0x800000)?(0xFF000000|PS_ctx.P_raw):PS_ctx.P_raw;
  PS_ctx.P_raw_sc = (float)PS_ctx.P_raw / PS_ctx.Kp;
  // PS_ctx.P_buffer = ((int64_t)PS_ctx.dig_C00 
                    // + (float)PS_ctx.P_raw * ((int64_t)PS_ctx.dig_C10 + (float)PS_ctx.P_raw * ((int64_t)PS_ctx.dig_C20 / (float)PS_ctx.Kp
                    // + (float)PS_ctx.P_raw * (int64_t)PS_ctx.dig_C30 / (float)PS_ctx.Kp)) / (float)PS_ctx.Kp 
                    // + (float)PS_ctx.T_raw * (int64_t)PS_ctx.dig_C01 / (float)PS_ctx.Kt 
                    // + PS_ctx.P_raw_sc * ((int64_t)PS_ctx.dig_C11 
                    // + (float)PS_ctx.P_raw * (int64_t)PS_ctx.dig_C21 / (float)PS_ctx.Kp) / (float)PS_ctx.Kp * (float)PS_ctx.T_raw / (float)PS_ctx.Kt);
  // PS_ctx.P_comp = (float)(PS_ctx.P_buffer);
  
  qua2 = PS_ctx.dig_C10  + PS_ctx.P_raw_sc * (PS_ctx.dig_C20  + PS_ctx.P_raw_sc* PS_ctx.dig_C30 );
  qua3 = PS_ctx.T_raw_sc * PS_ctx.P_raw_sc * (PS_ctx.dig_C11  + PS_ctx.P_raw_sc * PS_ctx.dig_C21 );
  PS_ctx.P_comp = PS_ctx.dig_C00  + PS_ctx.P_raw_sc * qua2 + PS_ctx.T_raw_sc * PS_ctx.dig_C01 + qua3;
  //(int32_t)(PS_ctx.P_comp/10);
  baroPressure = (int32_t)(PS_ctx.P_comp);
}

void init_AirPressure() {
	
  uint8_t buffer_reg = 0;
  delay(10);

  delay(250);
  delay(250);
  delay(250);
  delay(250);
  i2c_PS_SW_Reset();
  delay(100);


  while ((PS_ctx.Buff_status & PS_STATUS_SENSOR_RDY) == 0)
  {
    i2c_PS_ReadStatus();
    delay(100);
  }
  delay(250);
  while ((PS_ctx.Buff_status & PS_STATUS_COEF_RDY) == 0)
  {
    i2c_PS_ReadStatus();
    delay(100);
  }

  delay(250);
  i2c_PS_readCalibration();
  delay(10);

  //i2c_writeReg(PS_ADDRESS,PS_REGISTER_PSR_CFG,0b00000100);
  delay(10);

  PS_ctx.t_OSR = PM_PRC_2;
  PS_ctx.p_OSR = PM_PRC_8;

  i2c_writeReg(PS_ADDRESS, PS_REGISTER_PSR_CFG, PM_RATE_128 | PS_ctx.p_OSR);
  delay(10);

  i2c_writeReg(PS_ADDRESS, PS_REGISTER_TMP_CFG, TMP_RATE_32 | PS_ctx.t_OSR);
  delay(10);

  buffer_reg = FIFO_EN | INT_FIFO_FULL_CFG;
  if (PS_ctx.t_OSR > TMP_PRC_8) buffer_reg |= CFG_T_SHIFT;
  if (PS_ctx.p_OSR > PM_PRC_8) buffer_reg |= CFG_P_SHIFT;

  i2c_writeReg(PS_ADDRESS, PS_REGISTER_CFG_REG, buffer_reg);
  delay(10);

  if (PS_ctx.t_OSR == TMP_PRC_1) PS_ctx.Kt = 524288;    // 1 time
  else if (PS_ctx.t_OSR == TMP_PRC_2) PS_ctx.Kt = 1572864; // 2 time
  else if (PS_ctx.t_OSR == TMP_PRC_4) PS_ctx.Kt = 3670016; // 4 time
  else if (PS_ctx.t_OSR == TMP_PRC_8) PS_ctx.Kt = 7864320; // 8 time
  else if (PS_ctx.t_OSR == TMP_PRC_16) PS_ctx.Kt = 253952; // 16 time
  else if (PS_ctx.t_OSR == TMP_PRC_32) PS_ctx.Kt = 516096; // 32 time
  else if (PS_ctx.t_OSR == TMP_PRC_64) PS_ctx.Kt = 1040384; // 64 time
  else if (PS_ctx.t_OSR == TMP_PRC_128) PS_ctx.Kt = 2088960;             // 128 time

  if (PS_ctx.p_OSR == PM_PRC_1) PS_ctx.Kp = 524288;    // 1 time
  else if (PS_ctx.p_OSR == PM_PRC_2) PS_ctx.Kp = 1572864; // 2 time
  else if (PS_ctx.p_OSR == PM_PRC_4) PS_ctx.Kp = 3670016; // 4 time
  else if (PS_ctx.p_OSR == PM_PRC_8) PS_ctx.Kp = 7864320; // 8 time
  else if (PS_ctx.p_OSR == PM_PRC_16) PS_ctx.Kp = 253952; // 16 time
  else if (PS_ctx.p_OSR == PM_PRC_32) PS_ctx.Kp = 516096; // 32 time
  else if (PS_ctx.p_OSR == PM_PRC_64) PS_ctx.Kp = 1040384; // 64 time
  else if (PS_ctx.p_OSR == PM_PRC_128) PS_ctx.Kp = 2088960;             // 128 time
  
  PS_ctx.T_comp_LPF = 0;
  PS_ctx.P_comp_LPF = 0;

  delay(100);
  i2c_PS_MEAS_Start();
  //i2c_PS_UT_Start();

  //config_pressure(PM_RATE_128,PM_PRC_4);
  //delay(10);
  //config_temperature(TMP_RATE_128,TMP_PRC_2);
  //delay(10);
  //i2c_writeReg(PS_ADDRESS,PS_REGISTER_MEAS_CFG,MEAS_CTRL_ContinuousPressTemp); //????的???度?量
  //delay(20);

  //i2c_writeReg(PS_ADDRESS,PS_REGISTER_TMP_CFG,0b00000010);

  //i2c_writeReg(PS_ADDRESS,PS_REGISTE


  //delay(20);
  //i2c_PS_UT_Start();

  PS_ctx.deadline = currentTime + 10000;
  PS_ctx.comm_Error_cout = 0;
}

void config_pressure(uint8_t rate,uint8_t oversampling)
{
	PS_ctx.p_OSR = oversampling;
	
  if (PS_ctx.p_OSR == 0) PS_ctx.Kp = 524288;    // 1 time
  else if (PS_ctx.p_OSR == 1) PS_ctx.Kp = 1572864; // 2 time
  else if (PS_ctx.p_OSR == 2) PS_ctx.Kp = 3670016; // 4 time
  else if (PS_ctx.p_OSR == 3) PS_ctx.Kp = 7864320; // 8 time 
  else if (PS_ctx.p_OSR == 4) PS_ctx.Kp = 253952; // 16 time
  else if (PS_ctx.p_OSR == 5) PS_ctx.Kp = 516096; // 32 time
  else if (PS_ctx.p_OSR == 6) PS_ctx.Kp = 1040384; // 64 time
  else PS_ctx.Kp = 2088960;             // 128 time
  
	i2c_writeReg(PS_ADDRESS,PS_REGISTER_PSR_CFG,rate|oversampling);
}


void config_temperature(uint8_t rate,uint8_t oversampling)
{
	PS_ctx.t_OSR = oversampling;
	
  if (PS_ctx.t_OSR == 0) PS_ctx.Kt = 524288;    // 1 time
  else if (PS_ctx.t_OSR == 1) PS_ctx.Kt = 1572864; // 2 time
  else if (PS_ctx.t_OSR == 2) PS_ctx.Kt = 3670016; // 4 time
  else if (PS_ctx.t_OSR == 3) PS_ctx.Kt = 7864320; // 8 time
  else if (PS_ctx.t_OSR == 4) PS_ctx.Kt = 253952; // 16 time
  else if (PS_ctx.t_OSR == 5) PS_ctx.Kt = 516096; // 32 time
  else if (PS_ctx.t_OSR == 6) PS_ctx.Kt = 1040384; // 64 time
  else PS_ctx.Kt = 2088960;             // 128 time
  
	i2c_writeReg(PS_ADDRESS,PS_REGISTER_TMP_CFG,rate|oversampling);
}

//return 0: no data available, no computation ;  1: new value available  ; 2: no new value, but computation time
uint8_t Read_AirPressure() {                   // first UT conversion is started in init procedure
  /* if (currentTime < PS_ctx.deadline) return 0;
    else
    {
     i2c_PS_ReadStatus();
    if((PS_ctx.Buff_status & PS_STATUS_TMP_PRS_RDY) != PS_STATUS_TMP_PRS_RDY) return 0;

    PS_ctx.deadline = currentTime + 15000; // 1.5ms margin according to the spec (4.5ms T convetion time)
    i2c_PS_UT_Read_Command();
    i2c_PS_UT_Read();
    i2c_PS_UP_Read_Command();
    i2c_PS_UP_Read();
    i2c_PS_Calculate();
    return 1;
    } */
  if (currentTime < PS_ctx.deadline) return 0;
  else;
  
  i2c_PS_FIFO_Status();

  if ((PS_ctx.FIFO_status & 0x01) == 0)
  {
    PS_ctx.deadline = currentTime + 6000; // 1.5ms margin according to the spec (4.5ms T convetion time)
    i2c_PS_UP_Read_Command();
    i2c_PS_UP_Read();
    PS_ctx.comm_Error_cout = 0;
	if((PS_ctx.P_buffer[2] & 0x01) == 0)
	{
		PS_ctx.T_buffer[0] = PS_ctx.P_buffer[0];
		PS_ctx.T_buffer[1] = PS_ctx.P_buffer[1];
		PS_ctx.T_buffer[2] = PS_ctx.P_buffer[2];
		i2c_PS_Temp_Calculate();
		return 1;
	}
    else if((PS_ctx.P_buffer[2] & 0x01) == 1)
	{
		i2c_PS_Pressure_Calculate();
		return 2;
	}
  }
	return 1;
}


// ************************************************************************************************************
// I2C Gyroscope and Accelerometer MPU6050
// ************************************************************************************************************
#define MPU6050_ADDRESS     0x68 // address pin AD0 low (GND), default for FreeIMU v0.4 and InvenSense evaluation board
//#define MPU6050_ADDRESS     0x69 // address pin AD0 high (VCC)


void CodingDrone::MPU6050_GYRO_READ() {
  i2c_getSixRawADC(MPU6050_ADDRESS, 0x43);
  GYRO_ORIENTATION( ((rawADC[0]<<8) | rawADC[1])>>2 , // range: +/- 8192; +/- 2000 deg/sec
                    ((rawADC[2]<<8) | rawADC[3])>>2 ,
                    ((rawADC[4]<<8) | rawADC[5])>>2 );
  GYRO_Common();
}

void init_MPU6050 () {
 
  i2c_writeReg(MPU6050_ADDRESS, 0x6B, 0x80);             //PWR_MGMT_1    -- DEVICE_RESET 1
  delay(50);
  i2c_writeReg(MPU6050_ADDRESS, 0x6B, 0x03);             //PWR_MGMT_1    -- SLEEP 0; CYCLE 0; TEMP_DIS 0; CLKSEL 3 (PLL with Z Gyro reference)
  i2c_writeReg(MPU6050_ADDRESS, 0x1A, GYRO_DLPF_CFG);    //CONFIG        -- EXT_SYNC_SET 0 (disable input pin for data sync) ; default DLPF_CFG = 0 => ACC bandwidth = 260Hz  GYRO bandwidth = 256Hz)
  i2c_writeReg(MPU6050_ADDRESS, 0x1B, 0x18);             //GYRO_CONFIG   -- FS_SEL = 3: Full scale set to 2000 deg/sec
  i2c_writeReg(MPU6050_ADDRESS, 0x37, 0x02);           //INT_PIN_CFG   -- INT_LEVEL=0 ; INT_OPEN=0 ; LATCH_INT_EN=0 ; INT_RD_CLEAR=0 ; FSYNC_INT_LEVEL=0 ; FSYNC_INT_EN=0 ; I2C_BYPASS_EN=1 ; CLKOUT_EN=0
  i2c_writeReg(MPU6050_ADDRESS, 0x1C, 0x10);             //ACCEL_CONFIG  -- AFS_SEL=2 (Full Scale = +/-8G)  ; ACCELL_HPF=0   //note something is wrong in the spec.
  
}

void CodingDrone::MPU6050_ACC_READ () {
  i2c_getSixRawADC(MPU6050_ADDRESS, 0x3B);
  ACC_ORIENTATION( ((rawADC[0]<<8) | rawADC[1])>>3 ,
                   ((rawADC[2]<<8) | rawADC[3])>>3 ,
                   ((rawADC[4]<<8) | rawADC[5])>>3 );
  ACC_Common();
}


void initS() {
  i2c_init();
  init_MPU6050();
  init_AirPressure();
}

void CodingDrone::initSensors() {
	
  uint8_t c = 5;
  
  
  delay(10);
  
  while(c) { // We try several times to init all sensors without any i2c errors. An I2C error at this stage might results in a wrong sensor settings
    c--;
    initS();
    if (i2c_errors_count == 0) break; // no error during init => init ok
  }
  
  f.SMALL_ANGLES_25=1; // important for gyro only conf
  
  delay(10);
}



void CodingDrone::Attitude_by_MPU6050() {
  uint8_t axis;
  static int16_t gyroADCprevious[3] = {0,0,0};
  static int16_t gyroADCinter[3];

  uint16_t timeInterleave = 0;
  
  MPU6050_ACC_READ();
  if((calibratingA == 0 ) && (calibratingG == 0))getEstimatedAttitude();
  else {
	  
  }
  
  MPU6050_GYRO_READ();
  
  for (axis = 0; axis < 3; axis++)
    gyroADCinter[axis] =  imu.gyroADC[axis];

  timeInterleave=micros();
  
  uint8_t t=0;
  while((int16_t)(micros()-timeInterleave)<650) t=1; //empirical, interleaving delay between 2 consecutive reads
  
  MPU6050_GYRO_READ();
  
  for (axis = 0; axis < 3; axis++) {
    gyroADCinter[axis] =  imu.gyroADC[axis]+gyroADCinter[axis];
    // empirical, we take a weighted value of the current and the previous values
    imu.gyroData[axis] = (gyroADCinter[axis]+gyroADCprevious[axis])/3;
    gyroADCprevious[axis] = gyroADCinter[axis]>>1;
  }
  
  //magHold = att.heading;
}

// **************************************************
// Simplified IMU based on "Complementary Filter"
// Inspired by http://starlino.com/imu_guide.html
//
// adapted by ziss_dm : http://www.multiwii.com/forum/viewtopic.php?f=8&t=198
//
// The following ideas was used in this project:
// 1) Rotation matrix: http://en.wikipedia.org/wiki/Rotation_matrix
// 2) Small-angle approximation: http://en.wikipedia.org/wiki/Small-angle_approximation
// 3) C. Hastings approximation for atan2()
// 4) Optimization tricks: http://www.hackersdelight.org/
//
// Currently Magnetometer uses separate CF which is used only
// for heading approximation.
//
// **************************************************

//******  advanced users settings *******************
/* Set the Low Pass Filter factor for ACC
   Increasing this value would reduce ACC noise (visible in GUI), but would increase ACC lag time
   Comment this if  you do not want filter at all.
   unit = n power of 2 */
// this one is also used for ALT HOLD calculation, should not be changed
#ifndef ACC_LPF_FACTOR
  #define ACC_LPF_FACTOR 4 // that means a LPF of 16
#endif

/* Set the Gyro Weight for Gyro/Acc complementary filter
   Increasing this value would reduce and delay Acc influence on the output of the filter*/
#ifndef GYR_CMPF_FACTOR
  #define GYR_CMPF_FACTOR 10 //  that means a CMP_FACTOR of 1024 (2^10)
#endif

/* Set the Gyro Weight for Gyro/Magnetometer complementary filter
   Increasing this value would reduce and delay Magnetometer influence on the output of the filter*/
#define GYR_CMPFM_FACTOR 8 // that means a CMP_FACTOR of 256 (2^8)


typedef struct  {
  int32_t X,Y,Z;
} t_int32_t_vector_def;

typedef struct  {
  uint16_t XL; int16_t X;
  uint16_t YL; int16_t Y;
  uint16_t ZL; int16_t Z;
} t_int16_t_vector_def;

// note: we use implicit first 16 MSB bits 32 -> 16 cast. ie V32.X>>16 = V16.X
typedef union {
  int32_t A32[3];
  t_int32_t_vector_def V32;
  int16_t A16[6];
  t_int16_t_vector_def V16;
} t_int32_t_vector;

//return angle , unit: 1/10 degree
int16_t _atan2(int32_t y, int32_t x){
  float z = y;
  int16_t a;
  uint8_t c;
  c = abs(y) < abs(x);
  if ( c ) {z = z / x;} else {z = x / z;}
  a = 2046.43 * (z / (3.5714 +  z * z));
  if ( c ){
   if (x<0) {
     if (y<0) a -= 1800;
     else a += 1800;
   }
  } else {
    a = 900 - a;
    if (y<0) a -= 1800;
  }
  return a;
}

float fast_acos (float a)
{
    const float C  = 0.10501094f;
    float r, s, t, u;
    t = (a < 0) ? (-a) : a;  // handle negative arguments
    u = 1.0f - t;
    s = sqrtf (u + u);
    r = C * u * s + s;  // or fmaf (C * u, s, s) if FMA support in hardware
    if (a < 0) r = PI - r;  // handle negative arguments
    return r;
}

float InvSqrt (float x){ 
  union{  
    int32_t i;  
    float   f; 
  } conv; 
  conv.f = x; 
  conv.i = 0x5f1ffff9 - (conv.i >> 1); 
  return conv.f * (1.68191409f - 0.703952253f * x * conv.f * conv.f);
}

// signed16 * signed16
// 22 cycles
// http://mekonik.wordpress.com/2009/03/18/arduino-avr-gcc-multiplication/
#define MultiS16X16to32(longRes, intIn1, intIn2) \
asm volatile ( \
"clr r26 \n\t" \
"mul %A1, %A2 \n\t" \
"movw %A0, r0 \n\t" \
"muls %B1, %B2 \n\t" \
"movw %C0, r0 \n\t" \
"mulsu %B2, %A1 \n\t" \
"sbc %D0, r26 \n\t" \
"add %B0, r0 \n\t" \
"adc %C0, r1 \n\t" \
"adc %D0, r26 \n\t" \
"mulsu %B1, %A2 \n\t" \
"sbc %D0, r26 \n\t" \
"add %B0, r0 \n\t" \
"adc %C0, r1 \n\t" \
"adc %D0, r26 \n\t" \
"clr r1 \n\t" \
: \
"=&r" (longRes) \
: \
"a" (intIn1), \
"a" (intIn2) \
: \
"r26" \
)

int32_t  __attribute__ ((noinline)) mul(int16_t a, int16_t b) {
  int32_t r;
  MultiS16X16to32(r, a, b);
  //r = (int32_t)a*b; without asm requirement
  return r;
}

// Rotate Estimated vector(s) with small angle approximation, according to the gyro data
void rotateV32( t_int32_t_vector *v,int16_t* delta) {
  int16_t X = v->V16.X;
  int16_t Y = v->V16.Y;
  int16_t Z = v->V16.Z;

  v->V32.Z -=  mul(delta[ROLL]  ,  X)  + mul(delta[PITCH] , Y);
  v->V32.X +=  mul(delta[ROLL]  ,  Z)  - mul(delta[YAW]   , Y);
  v->V32.Y +=  mul(delta[PITCH] ,  Z)  + mul(delta[YAW]   , X);
}

static int16_t accZ=0;

void CodingDrone::getEstimatedAttitude(){
  uint8_t axis;
  int32_t accMag = 0;
  float scale;
  int16_t deltaGyroAngle16[3];
  static t_int32_t_vector EstG = {0,0,(int32_t)ACC_1G<<16};
  #if MAG
    static t_int32_t_vector EstM;
  #else
    static t_int32_t_vector EstM = {0,(int32_t)1<<24,0};
  #endif
  static uint32_t LPFAcc[3];
  float invG; // 1/|G|
  static int16_t accZoffset = 0;
  int32_t accZ_tmp=0;
  static uint16_t previousT;
  uint16_t currentT = micros();
  
  if(Reset_Att_Data_Flag){
	  Reset_Att_Data_Flag = 0;
	  accZoffset = 0;
	  
    EstG.A32[0] = 0;
	EstG.A32[1] = 0;
	EstG.A32[2] = (int32_t)ACC_1G<<16;
	
	LPFAcc[0] = 0;
	LPFAcc[1] = 0;
	LPFAcc[2] = 0;
	
  #if MAG
    EstM.A32[0] = 0;
    EstM.A32[1] = 0;
	EstM.A32[2] = 0;
  #else
    EstM.A32[0] = 0;
	EstM.A32[1] = (int32_t)1<<24;
	EstM.A32[2] = 0;
  #endif
  }

  // unit: radian per bit, scaled by 2^16 for further multiplication
  // with a delta time of 3000 us, and GYRO scale of most gyros, scale = a little bit less than 1
  scale = (currentT - previousT) * (GYRO_SCALE * 65536);
  previousT = currentT;

  // Initialization
  for (axis = 0; axis < 3; axis++) {
    // valid as long as LPF_FACTOR is less than 15
    imu.accSmooth[axis]  = LPFAcc[axis]>>ACC_LPF_FACTOR;
    LPFAcc[axis]      += imu.accADC[axis] - imu.accSmooth[axis];
    // used to calculate later the magnitude of acc vector
    accMag   += mul(imu.accSmooth[axis] , imu.accSmooth[axis]);
    // unit: radian scaled by 2^16
    // imu.gyroADC[axis] is 14 bit long, the scale factor ensure deltaGyroAngle16[axis] is still 14 bit long
    deltaGyroAngle16[axis] = imu.gyroADC[axis]  * scale;
  }

  // we rotate the intermediate 32 bit vector with the radian vector (deltaGyroAngle16), scaled by 2^16
  // however, only the first 16 MSB of the 32 bit vector is used to compute the result
  // it is ok to use this approximation as the 16 LSB are used only for the complementary filter part
  rotateV32(&EstG,deltaGyroAngle16);
  rotateV32(&EstM,deltaGyroAngle16);

  // Apply complimentary filter (Gyro drift correction)
  // If accel magnitude >1.15G or <0.85G and ACC vector outside of the limit range => we neutralize the effect of accelerometers in the angle estimation.
  // To do that, we just skip filter, as EstV already rotated by Gyro
  for (axis = 0; axis < 3; axis++) {
    if ( (int16_t)(0.85*ACC_1G*ACC_1G/256) < (int16_t)(accMag>>8) && (int16_t)(accMag>>8) < (int16_t)(1.15*ACC_1G*ACC_1G/256) )
      EstG.A32[axis] += (int32_t)(imu.accSmooth[axis] - EstG.A16[2*axis+1])<<(16-GYR_CMPF_FACTOR);
    accZ_tmp += mul(imu.accSmooth[axis] , EstG.A16[2*axis+1]);
    #if MAG
      EstM.A32[axis]  += (int32_t)(imu.magADC[axis] - EstM.A16[2*axis+1])<<(16-GYR_CMPFM_FACTOR);
    #endif
  }
  
  
  if (rcOptions[BOXARM] == 0 && f.ARMED) {
	  if(imu.accSmooth[YAW] > 700) Impact_Flag = 1;
  }
  
  
  //att.ACCZ_Feedback = imu.accSmooth[YAW]-512;
  //att.ACCZ_Feedback = imu.accSmooth[YAW]-512;
  //att.Angle_Z = (float)imu.accSmooth[YAW]/512;
  //att.Angle_Feedback = fast_acos(att.Angle_Z);
  //att.Angle_Feedback = (float)512/imu.accSmooth[YAW];
  //att.Angle_Z = 512 - imu.accSmooth[YAW];
  
  if (EstG.V16.Z < ACCZ_25deg)
    f.SMALL_ANGLES_25 = 1;
  else
    f.SMALL_ANGLES_25 = 0;


  if(f.ARMED) {
	  if(EstG.V16.Z < ACCZ_Failsafe_deg) FailSafe_Flag = 1;
  }
  else FailSafe_Flag = 0;

  // Attitude of the estimated vector
  int32_t sqGX_sqGZ = mul(EstG.V16.X,EstG.V16.X) + mul(EstG.V16.Z,EstG.V16.Z);
  invG = InvSqrt(sqGX_sqGZ + mul(EstG.V16.Y,EstG.V16.Y));
  att.angle[ROLL]  = _atan2(EstG.V16.X , EstG.V16.Z);
  att.angle[PITCH] = _atan2(EstG.V16.Y , InvSqrt(sqGX_sqGZ)*sqGX_sqGZ);

  //note on the second term: mathematically there is a risk of overflow (16*16*16=48 bits). assumed to be null with real values
  att.heading = _atan2(
    mul(EstM.V16.Z , EstG.V16.X) - mul(EstM.V16.X , EstG.V16.Z),
    (EstM.V16.Y * sqGX_sqGZ  - (mul(EstM.V16.X , EstG.V16.X) + mul(EstM.V16.Z , EstG.V16.Z)) * EstG.V16.Y)*invG );
  
  //att.heading /= 10;
  
  #if defined(THROTTLE_ANGLE_CORRECTION)
    cosZ = mul(EstG.V16.Z , 100) / ACC_1G ;                                                   // cos(angleZ) * 100 
    throttleAngleCorrection = THROTTLE_ANGLE_CORRECTION * constrain(100 - cosZ, 0, 100) >>3;  // 16 bit ok: 200*150 = 30000  
  #endif
  
  
  
  // projection of ACC vector to global Z, with 1G subtructed
  // Math: accZ = A * G / |G| - 1G
  accZ = accZ_tmp *  invG;
  if (!f.ARMED) {
    accZoffset -= accZoffset>>3;
    accZoffset += accZ;
  }  
  accZ -= accZoffset>>3;
}


int CodingDrone::angle_ROLL(){
  return (int)(att.angle[0])/10;
}
int CodingDrone::angle_PITCH(){
  return (int)(att.angle[1])/10;
}
int CodingDrone::angle_YAW(){
  return (int)(att.heading)/10;
}
int CodingDrone::ALTITUDE_Data(){
  return (int)(alt.EstAlt);
}

int CodingDrone::Sensor_Data(uint8_t addr){
	switch(addr)
	{
		case 0 : 
		case 1 : return (int)(att.angle[addr])/10;
			break;
		case 2 : return (int)(att.heading)/10;
			break;
		case 3 : return (int)(alt.EstAlt);
			break;
	}
}


uint8_t getEstimatedAltitude(){
  float  BaroAlt = 0;
  static float baroGroundTemperatureScale,logBaroGroundPressureSum;
  static float vel = 0.0f;
  static uint16_t previousT;
  uint16_t currentT = micros();
  uint16_t dTime;
  static int16_t Compansation = 0;
  static float lastBaroAlt = 0;
  static float errorAltitudeI = 0;

  dTime = currentT - previousT;
  if (dTime < UPDATE_INTERVAL) return 0;
  previousT = currentT;

  if(calibratingB > 0) {
	logBaroGroundPressureSum = 4433076 * (1 - pow((PS_ctx.P_comp/101325), 0.190263));
    //baroGroundTemperatureScale = (PS_ctx.t_comp + 27315) * 2 * (29.271267f); // 2 *  is included here => no need for * 2  on BaroAlt in additional LPF
    calibratingB--;
  }
  // baroGroundPressureSum is not supposed to be 0 here
  // see: https://code.google.com/p/ardupilot-mega/source/browse/libraries/AP_Baro/AP_Baro.cpp
  BaroAlt = 4433076 * (1 - pow((PS_ctx.P_comp/101325), 0.190263)) - logBaroGroundPressureSum;
  //BaroAlt = (logBaroGroundPressureSum - log(baroPressureSum));
  
  
  float error16 = 0;
	  
  //alt.EstAlt = BaroAlt/2;//(alt.EstAlt * 6 + BaroAlt ) >> 3; // additional LPF to reduce baro noise (faster by 30 µs)
  //alt.EstAlt = (alt.EstAlt * 7 + BaroAlt ) >> 3; // additional LPF to reduce baro noise (faster by 30 µs)
  //if(f.ARMED) alt.EstAlt = ((alt.EstAlt * 7 + BaroAlt - (rcCommand[THROTTLE] - 1000 + 300 - Alt_Throttle_Compensation)*11/100) >> 3);
  if(f.ARMED) {
	  //alt.EstAlt = ((alt.EstAlt * 7 + BaroAlt - ((((motor[0]+motor[1]+motor[2]+motor[3])>>2) - 1000) + 300 - Alt_Throttle_Compensation)/11) / 8);
	  
	  alt.EstAlt = (alt.EstAlt * 7 + BaroAlt ) / 8;
	
  
	  //if(alt.direction == 'N') error16 = constrain(AltHold - alt.EstAlt + ((512-att.Angle_Feedback)/4) , -300, 300);
	  //else;
	  #if defined(THROTTLE_ANGLE_ALTITUDE_COMP)
		error16 = constrain(AltHold - alt.EstAlt + abs(att.angle[ROLL]/8) + abs(att.angle[PITCH]/8), -500, 500);
	  #else
		error16 = constrain((AltHold - alt.EstAlt)*4, -1000, 1000);	
	  #endif
	  
	  //P 
	  applyDeadband(error16, 4); //remove small P parametr to reduce noise near zero position
	  BaroPID = constrain((conf.pid[PIDALT].P8 * error16 / 128), -500, 500);

	  //I
	  errorAltitudeI += conf.pid[PIDALT].I8 * error16 / 100;
	  errorAltitudeI = constrain(errorAltitudeI,-16000,30000);
	  BaroPID += errorAltitudeI/500; //I in range +/-60
	 
	  //applyDeadband(accZ, ACC_Z_DEADBAND);

	  // could only overflow swith a difference of 320m, which is highly improbable here
	  int16_t baroVel = mul((alt.EstAlt - lastBaroAlt) , (1000000 / UPDATE_INTERVAL));

	  lastBaroAlt = alt.EstAlt;
	  

	  baroVel = constrain(baroVel, -300, 300); // constrain baro velocity +/- 300cm/s
	  applyDeadband(baroVel, 5); // to reduce noise near zero

	  // Integrator - velocity, cm/sec
	  
	  vel += accZ * ACC_VelScale * dTime;
	  
	  //constrain(conf.pid[PIDALT].D8*cosZ/512, 10, 20);
	  
	  // apply Complimentary Filter to keep the calculated velocity based on baro velocity (i.e. near real velocity). 
	  // By using CF it's possible to correct the drift of integrated accZ (velocity) without loosing the phase, i.e without delay
	  vel = vel * 0.985f + baroVel * 0.015f;
	  //vel = baroVel;
	  
	  //D
	  alt.vario = vel;
	  applyDeadband(alt.vario, 5);
	  BaroPID -= constrain(conf.pid[PIDALT].D8 * alt.vario / 50, -1500, 1500);
  }
  else {
	  alt.EstAlt = (alt.EstAlt * 7 + BaroAlt ) / 8;
	  lastBaroAlt = alt.EstAlt;
	  vel = 0;
	  errorAltitudeI = 0;
	  BaroPID = 0;
  }
  
  //alt.EstAlt = (alt.EstAlt * 7 + BaroAlt ) / 8;
  
  
  return 1;
  
}


